/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/


/**
 * @file util/parseconfig.c
 * @brief configuration file parsing 
 * @author Gerd Knorr <kraxel@bytesex.org> 
 * 
 * Configuration file parsing, taken from xawtv (GPL).
 * Some adjustments and dead code elimination by Christian Grothoff.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
//#include "util/parseconfig.h"
#include "util/logging.h"
#include "util/xmalloc.h"

struct CFG_ENTRIES {
    int  ent_count;
    char **ent_names;
    char **ent_values;
};

struct CFG_SECTIONS {
    int                 sec_count;
    char                **sec_names;
    struct CFG_ENTRIES  **sec_entries;
};

/* ------------------------------------------------------------------------ */

static struct CFG_SECTIONS * c = NULL;

/* ------------------------------------------------------------------------ */

#define ALLOC_SIZE 16


static struct CFG_SECTIONS * cfg_init_sections(void) {
  struct CFG_SECTIONS *c;
  c = MALLOC(sizeof(struct CFG_SECTIONS));
  memset(c,
	 0,
	 sizeof(struct CFG_SECTIONS));
  c->sec_names   = MALLOC(ALLOC_SIZE*sizeof(char*));
  c->sec_entries = MALLOC(ALLOC_SIZE*sizeof(struct CFG_ENTRIES*));
  return c;
}

static struct CFG_ENTRIES * cfg_init_entries() {
  struct CFG_ENTRIES *e;
  e = MALLOC(sizeof(struct CFG_ENTRIES));
  memset(e,
	 0,
	 sizeof(struct CFG_ENTRIES));
  e->ent_names  = MALLOC(ALLOC_SIZE*sizeof(char*));
  e->ent_values = MALLOC(ALLOC_SIZE*sizeof(char*));
  return e;
}

static struct CFG_ENTRIES * cfg_find_section(struct CFG_SECTIONS *c, 
					     char * name) {
  struct CFG_ENTRIES * e;
  int i;
  
  for (i=0; i<c->sec_count; i++)
    if (0 == strcasecmp(c->sec_names[i], name))
      return c->sec_entries[i];
  
  /* 404 not found => create a new one */
  if ((c->sec_count % ALLOC_SIZE) == (ALLOC_SIZE-1)) {
    i = c->sec_count+1;
    GROW(c->sec_names,
	 i,
	 i+ALLOC_SIZE);
    i = c->sec_count+1;
    GROW(c->sec_entries,
	 i,
	 i+ALLOC_SIZE);
  }
  e = cfg_init_entries();
  c->sec_names[c->sec_count]   = STRDUP(name);
  c->sec_entries[c->sec_count] = e;
  c->sec_count++;    
  return e;
}

static void cfg_set_entry(struct CFG_ENTRIES * e, 
			  char * name, 
			  char * value) {
  int i;
  
  for (i=0; i<e->ent_count; i++)
    if (0 == strcasecmp(e->ent_names[i],
			name))
      break;
  if (i == e->ent_count) {
    /* 404 not found => create a new one */
    if ((e->ent_count % ALLOC_SIZE) == (ALLOC_SIZE-1)) {
      /* not enough space, grow first */
      i = e->ent_count+1;
      GROW(e->ent_names,
	   i,
	   i+ALLOC_SIZE);
      i = e->ent_count+1;
      GROW(e->ent_values,
	   i,
	   i+ALLOC_SIZE);
    }
    i = e->ent_count;
    e->ent_count++;    
  } else {
    /* free old values, will be replaced! */
    FREENONNULL(e->ent_names[i]);
    FREENONNULL(e->ent_values[i]);
  }      
  e->ent_names[i]  = STRDUP(name);
  e->ent_values[i] = STRDUP(value);
}

/* ------------------------------------------------------------------------ */

int
cfg_parse_file(char *filename)
{
    struct CFG_ENTRIES * e = NULL;
    char line[256],tag[64],value[192];
    FILE *fp;
    int nr;
    int i;
    int emptyline;
    
    if (NULL == c)
	c = cfg_init_sections();
    if (NULL == (fp = fopen(filename,"r")))
	return -1;

    memset(line, 
           0, 
	   256);

    nr = 0;
    while (NULL != fgets(line,255,fp)) {
	nr++;
	for (i=0;i<255;i++) {
	  if (line[i] == '\t')
	    line[i] = ' ';
	}
	emptyline=1;
	for (i=0;(i<255 && line[i] != 0);i++) {
	  if (line[i] != ' ' && line[i] != '\n')
	    emptyline=0;
	}
	if (emptyline == 1)
	  continue;
	if (line[0] == '\n' || line[0] == '#' || line[0] == '%')
	    continue;
	for (i=strlen(line)-2 ; 
	     (i>=0) && (line[i] == ' ' || line[i] == '\t' ) ;
	     i--)
	    line[i] = 0;
	if (1 == sscanf(line,"[%99[^]]]", value)) {
	    /* [section] */
	    e = cfg_find_section(c,value);
	} else if (2 == sscanf(line," %63[^= ] = %191[^\n]",tag,value)) {
	    /* foo = bar */
	  if (NULL == e) /* no section defined so far: put in "global" section (change by CG) */
	    e = cfg_find_section(c, "");  
	  i=0;
	  if (value[0] == '"') {
	    i=1;
	    while ( (value[i] != '\0') && 
		    (value[i] != '"') )
	      i++;
	    if (value[i] == '"') {
	      value[i] = '\0';
	      i=1;
	    } else
	      i=0;
	  }
	  cfg_set_entry(e, tag, &value[i]);
	    
	} else {
	    /* Huh ? */
	  LOG(LOG_ERROR,
	      "ERROR: %s:%d: syntax error\n",
	      filename,
	      nr);
	}
    }
    fclose(fp);
    return 0;
}

/* ------------------------------------------------------------------------ */

char * cfg_get_str(char * sec, 
		   char * ent) {
  struct CFG_ENTRIES * e = NULL;
  int i;
  
  for (i = 0; i < c->sec_count; i++)
    if (0 == strcasecmp(c->sec_names[i],sec))
      e = c->sec_entries[i];
  if (NULL == e)
    return NULL;
  for (i = 0; i < e->ent_count; i++)
    if (0 == strcasecmp(e->ent_names[i],ent)) {
      return e->ent_values[i];
    }
  return NULL;
}

int cfg_get_signed_int(char *sec, 
		       char *ent) {
  char *val;

  val = cfg_get_str(sec, ent);    
  if (NULL == val)
    return 0;
  return atoi(val);
}

void doneParseConfig() {
  int i;
  int j;

  if (c == NULL)
    return;	      
  for (i=0;i<c->sec_count;i++) {
    if (c->sec_entries[i] != NULL) {
      for (j=0;j<c->sec_entries[i]->ent_count;j++) {
	FREENONNULL(c->sec_entries[i]->ent_names[j]);
	FREENONNULL(c->sec_entries[i]->ent_values[j]);
      }    
      FREENONNULL(c->sec_entries[i]->ent_names);
      FREENONNULL(c->sec_entries[i]->ent_values);
    }
    FREENONNULL(c->sec_entries[i]);
    FREENONNULL(c->sec_names[i]);
  }
  FREENONNULL(c->sec_entries);
  FREENONNULL(c->sec_names);
  FREENONNULL(c);
  c = NULL;
}

/* end of parseconfig.c */
